package frc.robot.subsystems;

import java.util.function.DoubleSupplier;

import com.revrobotics.CANSparkMax;
import com.revrobotics.RelativeEncoder;
import com.revrobotics.CANSparkLowLevel.MotorType;

import edu.wpi.first.math.controller.PIDController;
import edu.wpi.first.wpilibj.ADXRS450_Gyro;
import edu.wpi.first.wpilibj.drive.DifferentialDrive;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import frc.robot.constants.DrivetrainConstants;

public class Drivetrain extends SubsystemBase {
    //creates instance variables//
    private CANSparkMax leftFront;
    private CANSparkMax leftRear;
    private CANSparkMax rightFront;
    private CANSparkMax rightRear;

    private RelativeEncoder leftFrontEncoder;
    private RelativeEncoder rightFrontEncoder;

    private ADXRS450_Gyro gyro;

    private PIDController turnController;
    private PIDController rightDistanceController;
    private PIDController leftDistanceController;
    
    private DifferentialDrive drive;

    public Drivetrain() {
        //creates instances of each motor//
        leftFront = new CANSparkMax(DrivetrainConstants.kLeftFrontCAN, MotorType.kBrushless);
        leftRear = new CANSparkMax(DrivetrainConstants.kLeftRearCAN, MotorType.kBrushless);
        rightFront = new CANSparkMax(DrivetrainConstants.kRightFrontCAN, MotorType.kBrushless);
        rightRear = new CANSparkMax(DrivetrainConstants.kRightRearCAN, MotorType.kBrushless);

        //makes the other motor in a gearbox follow its companion//
        leftRear.follow(leftFront);
        rightRear.follow(rightFront);
        rightFront.setInverted(true); //makes the right side of the drive base inverted//

        leftFrontEncoder = leftFront.getEncoder();
        rightFrontEncoder = rightFront.getEncoder();

        gyro = new ADXRS450_Gyro();
        gyro.calibrate();

        turnController = new PIDController(
            DrivetrainConstants.kTurnControllerP, 
            DrivetrainConstants.kTurnControllerI, 
            DrivetrainConstants.kTurnControllerD
        );
        turnController.setTolerance(DrivetrainConstants.kTurnControllerTolerance);

        rightDistanceController = new PIDController(
            DrivetrainConstants.kDistanceControllerP, 
            DrivetrainConstants.kDistanceControllerI, 
            DrivetrainConstants.kDistanceControllerD
        );
        rightDistanceController.setTolerance(DrivetrainConstants.kDistanceControllerTolerance);

        leftDistanceController = new PIDController(
            DrivetrainConstants.kDistanceControllerP,
            DrivetrainConstants.kDistanceControllerI,
            DrivetrainConstants.kDistanceControllerD
        );
        leftDistanceController.setTolerance(DrivetrainConstants.kDistanceControllerTolerance);

        drive = new DifferentialDrive(leftFront, rightFront);
    }

    //gyro methods//
    public double getGyroAngle() {
        return gyro.getAngle() * (DrivetrainConstants.kInvertGyro ? -1 : 1);
    }

    public void resetGyro() {
        gyro.reset();
        return;
    }

    //encoder methods//
    public double getLeftEncoderPos() {
        return leftFrontEncoder.getPosition();
    }

    public double getRightEncoderPos() {
        return rightFrontEncoder.getPosition();
    }

    public void resetDriveEncoders() {
        leftFrontEncoder.setPosition(0);
        rightFrontEncoder.setPosition(0);
        return;
    }
    
    //method to turn drivetrain a certain angle//
    public Command goToAngle(double angle, double timeout) {
        return run(() -> {
            double output = turnController.calculate(
                getGyroAngle(), 
                angle
            );
            drive.tankDrive(-output, output);
        }).until(turnController::atSetpoint).withTimeout(timeout);
    }

    //method to move drivetrain a certain distance//
    public Command moveDistance(double distance, double timeout) {
        return run(() -> {
            double outputL = rightDistanceController.calculate(
                (leftFrontEncoder.getPosition() + rightFrontEncoder.getPosition()) / 2, 
                distance
            );
            double outputR = rightDistanceController.calculate(
                (leftFrontEncoder.getPosition() + rightFrontEncoder.getPosition()) / 2, 
                distance
            );
            drive.tankDrive(outputL, outputR);
        }).until(rightDistanceController::atSetpoint).withTimeout(timeout);
    }

    //methods that take 2 doubles and convert them to motor commands//
    public Command driveArcade(DoubleSupplier xSpeed, DoubleSupplier zRotation) {
        return run(() -> {
            drive.arcadeDrive(xSpeed.getAsDouble(), zRotation.getAsDouble());
        });
    }

    public Command driveTank(DoubleSupplier leftSpeed, DoubleSupplier rightSpeed) {
        return run(() -> {
            drive.tankDrive(leftSpeed.getAsDouble(), rightSpeed.getAsDouble());
        });
    }
}